#ifndef TRXVOL_H
#define TRXVOL_H
/** \file
   Volume and surface definition.
*/
/*

   Copyright  1992-2010 M.Kraemer, GSI Darmstadt 

   $Log: trxvol.h,v $
   Revision 1.5  2009/05/29 12:34:30  kraemer
   reworked projectiles

   Revision 1.4  2008/06/20 14:29:37  kraemer
   tilted vol's

   Revision 1.3  2007/08/30 14:35:11  kraemer
   minor bug fixes / changes

   Revision 1.2  2002/11/25 16:38:48  kraemer
   surface/vol; volume mode; 2file

   Revision 1.1  2001/01/09 14:54:43  kraemer
   *** empty log message ***


*/

#include "cmderr.h"
#include "trxproj.h"
#include "trxxs.h"
#include "trxemfld.h"
                                     /* Common status and action flags */
                                      /* masks */
#define TRXVOL_MASKGLOBAL 0xFF000000  /* global (e.g. debug) */
#define TRXVOL_MASKFUNC   0x00FF0000  /* function specific flags */
#define TRXVOL_MASKQUAL   0x0000FFFF  /* for additional qualifiers */

                                      /* flags */
#define TRXVOL_DEBUG      0x80000000  /* general check flag */
#define TRXVOL_CHECK      TRXVOL_DEBUG/* general check flag */
#define TRXVOL_QUIET      0x10000000  /* suppress error messages */
#define TRXVOL_NOMSG      TRXVOL_QUIET

/*
   --- surface definitions
*/
                                      /** \struct STRXSRF
                                          \brief Surface description.
                                      */ 
struct STRXSRF {
   unsigned int uiFlag;              /**< flag bits */
   unsigned int uiShape;             /**< shape */
   unsigned int uiTopo;              /**< upper/lower  */
   unsigned int uiSN;                /**< neighbouring surface ID (1..n) */
   struct STRXVOL *psVN;              /**< neighbouring volume */
   double dvNorm[3];                  /**< normal vector */
};
                                      /* anti-surface */
#define TRXSRF_ANTI 0x00000001

                                      /* topology */
enum {
   TRXSRF_LOWER=0, TRXSRF_UPPER,
};                                          
                                      /* surface shapes */
enum {
   TRXSRF_CYL=1, TRXSRF_CIRCLE, TRXSRF_SPHERE, TRXSRF_RECT, TRXSRF_TORUS,
};

                                      /** \struct STRXVOLSCORE
                                          \brief Volume scoring.
                                      */ 
struct STRXVOLSCORE {
   unsigned int uivCount[TRXPROJ_MAX][TRXXS_MAXPSEUDOXS];  /**< count # of interactions */  
   double        dvELoss [TRXPROJ_MAX][TRXXS_MAXPSEUDOXS]; /**< count energy losses [keV] */ 
   double dEDeposit;                                    /**< deposited energy [keV], for a single primary event */ 
   double dWValue;                                      /**< W-value [keV] */ 
   int nEvt;                                            /**< # of events to be scored */
   unsigned int *puiFano;                               /**< [nEvt] # of ionizations per event, for Fano factor */ 
};                                    
                                       
/*
   --- volume definitions
*/
                                      /* data offsets */
enum {
   TRXVOL_HEIGHT=0, 
   TRXVOL_XMIN, TRXVOL_XMAX, 
   TRXVOL_YMIN, TRXVOL_YMAX, 
   TRXVOL_ZMIN, TRXVOL_ZMAX, 
   TRXVOL_RMIN, TRXVOL_RMAX, 
   TRXVOL_R2MIN, TRXVOL_R2MAX, 
   TRXVOL_THMIN, TRXVOL_THMAX,
   TRXVOL_PHMIN, TRXVOL_PHMAX,
   TRXVOL_DX, TRXVOL_DY, TRXVOL_DZ,
   TRXVOL_THICKNESS,
   MTRXVOLDATA 
};
                                      /** \struct STRXVOL
                                          \brief Volume description.
                                      */ 
struct STRXVOL {
   unsigned int uiFlag;               /**< flag bits */
   unsigned int  uiID;                /**< vol number, for reference */
   unsigned int uiShape;              /**< shape */
   size_t ulSpace;                    /**< occupied space (bytes) */
   double dVol;                       /**< volume [cm**3] */
   double dSrf;                       /**< surface [cm**2] */  
   double dMinDim;                    /**< smallest dimension [cm] */
   double dDensity;                   /**< local density [g/cm**3] (deviating from material) */ 
   double dvCutoff[TRXPROJ_MAX];         /**< low energy cutoff limit [keV or MeV/u] */  
   double dvData[MTRXVOLDATA];        /**< volume limits data */
   double dvCentre[3];                /**< centre coordinates [cm] */
   double dvWorld2Vol[4][4];          /**< transformation matrix World => Volume */
   double dvVol2World[4][4];          /**< transformation matrix Volume => World */
   double dvRot[3];                   /**< rotation angle [deg] x,y,z */
   int    ivRot[3];                   /**< rotation axis number (0,1,2 for x,y,z), in the order of application */
   int    nRot;                       /**< # of actual rotations */
   void *pvUser;                      /**< reserved for user info */
   char *pcName;                      /**< volume  name */
   char *pcMaterial;                  /**< material name */

#define MTRXVOLSRF 6
   int nS, mS;                        /**< current, max # of surfaces */
   struct STRXSRF sS[MTRXVOLSRF];     /**< surfaces */
   struct STRXVOLSCORE sScore;        /**< total score */
   struct STRXVOLSCORE sScorePrim;    /**< score for primary particles */
   struct STRXVOLSCORE sScoreSec;     /**< score for secondary particles */
   struct STRXEMFLD sE;               /**< electric field spec.s */
   struct STRXEMFLD sB;               /**< magnetic field spec.s */
   struct SCMDERR sCE;                /**< error handling */
};
                                      /* flag bits */
#define TRXVOL_RECORD        0x00000001  /* */
#define TRXVOL_REFCOPY       0x00000002  /* don't free() */

                                      /* volume shapes */
enum {
   TRXVOL_NONE=0,
   TRXVOL_CYL=1, TRXVOL_SPHERE, TRXVOL_BOX, TRXVOL_TORUS,
};


/*

   General:
   --------

   o where possible:
     surfaces w/ normal vectors, else n=0
     pointing outside volume,
     when intersection from outside: 
     dot product > 0: no intersection possible  


   Functions for single volumes: 
   -----------------------------
                            
   TRXVolComplete     ( vol, flag ) 
                      o derive quantities to complete vol description
                    5 x to test
                      o to complete
                        o shapes
                          o torus
                          o surface description for box ?
                        . trafo for tilted
                        o angular limits
                      o calls
  
   TRXVolCopy         ( volfr, volto, flag )
                      o reference copy, don't free !
                    4 x to test
                      o calls
                        o TRXVolInit()
 
   TRXVolFree         ( vol, flag ) 
                      o release resources 
                    3 x to test
                      o calls  
                        o TRXVolInit()
                        o TRXVolScoreFanoFree()
                        o TRXEFldFree()
                        o TRXBFldFree()

   TRXVolInit         ( vol, flag ) 
                      o initialize 
                    2 x to test
                      o calls
                        o TRXEFldInit()
                        o TRXBFldInit()
  
   TRXVolInspect      ( vol, flag ) 
                      o inspect output
                    1 x to test
                      o calls
                        o TRXEFldInspect()
                        o TRXBFldInspect()
  
   TRXVolIsValid      ( vol, flag ) 
                      o check if volume is minimum valid 
                    6 x to test
                      o to complete:
                        o shapes
                          o torus
                        ? valid rotation matrix
                        o angular limits
                        ? surfaces
                      o calls
  
   TRXVolSpace        ( vol, flag ) 
                      o count occupied space 
                   10 x to test
                      o calls
                        o TRXEMFldSpace() 
  
   TRXVolVolMatch     ( vol, vol, flag ) 
                      o check volume/surface natch
                    9 o to complete
                      ! tilted volumes (axis)
                      o sidewards checks
                      o angular checks
                      o sphere
                      o box
                      o torus
                      o calls
  
   TRXVol2File        ( vol, fp, flag) 
                      o output volume data to file, native
                    7 o to test
                      o to complete:
                        o shapes
                          o torus
                        x rotation
                        o angular limits
                      o calls
                        o TRXVol*()
                        o TRXEFld2File()
                        o TRXEFld2File()
                                                   
   TRXVol2Filegd      ( vol, fp, flag )
                      o output volume data to file, gd format
                    8 o to do
                      o to complete:
                        o shapes
                          o torus
                        o orientation
                        o angular limits
                      o calls 
                        o TRXVolIsValid()


   Functions for volume scoring: 
   -----------------------------

   TRXVolScoreEvaluate   ( vol, nevt, flag )
                       4 x to test
                         o evaluate/list score counters
                     			 ! density per material
                         o calls 

   TRXVolScoreFanoAlloc  ( vol, nevt, flag  ) 
                       3 x to test
                         o allocate Fano recording array
                         x recording vol only
                         o calls  

   TRXVolScoreFanoFree   ( vol, flag ) 
                       2 x to test
                         o release Fano resources  
                         o calls  

   TRXVolScoreReset      ( vol, flag )
                       1 x to test
                         o calls 


   Other functions
   ---------------

   TRXVolError        error handling
 
*/
                                      /* function prototypes */
int TRXVolComplete( struct STRXVOL *, unsigned int );
int TRXVolCopy    ( struct STRXVOL *, struct STRXVOL *, unsigned int );
int TRXVolFree    ( struct STRXVOL *, unsigned int );
int TRXVolInit    ( struct STRXVOL *, unsigned int );
int TRXVolInspect ( struct STRXVOL *, unsigned int );
int TRXVolIsValid ( struct STRXVOL *, unsigned int );
int TRXVolSpace   ( struct STRXVOL *, unsigned int );
int TRXVolVolMatch( struct STRXVOL *, struct STRXVOL *, unsigned int );
int TRXVol2Filegd ( struct STRXVOL *, FILE *, unsigned int );
int TRXVol2File   ( struct STRXVOL *, FILE *, unsigned int );

enum { 
   TRXVOLTEST_CYL1=0, TRXVOLTEST_CYL2,
   TRXVOLTEST_BOX1, TRXVOLTEST_BOX2,
   TRXVOLTEST_SPHERE1, TRXVOLTEST_SPHERE2,
   TRXVOLTEST_ROT,
   TRXVOLTEST_MAX,
};

struct STRXVOL *_TRXVolTest( unsigned int, unsigned int );

int TRXVolScoreEvaluate ( struct STRXVOL *, unsigned int, unsigned int );
int TRXVolScoreFanoAlloc( struct STRXVOL *, unsigned int, unsigned int );
int TRXVolScoreFanoFree ( struct STRXVOL *, unsigned int );
int TRXVolScoreReset    ( struct STRXVOL *, unsigned int );



#define TRXVOLEVT_FILEVERSION 20020830
                                      /** \struct STRXVOLEVTDATA
                                          \brief Volume event data.
                                      */ 
struct STRXVOLEVTDATA {
   uint32_t ui32Vol;                    /**< Vol ID */
   float           rdE;               /**< energy loss [keV] */
   float           rt;                /**< time [sec] */
};
                                      /** \struct STRXVOLEVTHEAD
                                          \brief Volume event header.
                                      */ 
struct STRXVOLEVTHEAD {
   uint32_t ui32Version, ui32Length;      /**< version, length of header+data */
   uint16_t ui16ByteSex;                /**< =1 for native endian, else byte swap */
   uint16_t ui16Mode;                   /**< always=1 for volmode evt */
   uint32_t ui32Evt;                    /**< current/max primary event number */
    int32_t i32OR;                      /**< BOR/EOR */
   int nD;                            /**< # of volevt data following */
};
                                      /** \struct STRXVOLEVT
                                          \brief Volume event output.
                                      */ 
struct STRXVOLEVT {
   struct STRXVOLEVTHEAD   sH;	       /**< Volume event header */   
   struct STRXVOLEVTDATA *psD;	       /**< Volume event data, to be written after header */
};
#define TRXVOLEVTFREE(psVE)\
   if ( (psVE)->psD != NULL ) free((psVE)->psD); memset(psVE,0,sizeof(*(psVE))); 
/*

   Volume event data functions
   ---------------------------
   
   TRXVolEvtInit     ( ve, flag )
                     o initialize
                     o to test
                     o calls
   
   TRXVolEvtInspect  ( ve, flag )
                     o inspect
                     o to test
                     o calls
   

*/
int TRXVolEvtInit   ( struct STRXVOLEVT *, unsigned int );
int TRXVolEvtInspect( struct STRXVOLEVT *, unsigned int );

struct STRXVOLEVT *_TRXVolEvtTest( unsigned int );
 
 

                                      /* error handling */
int TRXVolError( struct SCMDERR *, unsigned int );

enum { ERRTRXVOL_SYS=1,
       ERRTRXVOL_BADADDR,
       ERRTRXVOL_BADNEVT,
       ERRTRXVOL_NOSUCHSHAPE,
       ERRTRXVOL_BADVOLTYPE,
       ERRTRXVOL_BADVOLNAME,
       ERRTRXVOL_BADVOLMATERIAL,
       ERRTRXVOL_BADHEIGHT,
       ERRTRXVOL_BADRADIUS,
       ERRTRXVOL_BADANGLE,
       ERRTRXVOL_BADSIZE,
       ERRTRXVOL_BADTHICKNESS,
       ERRTRXVOL_BADROTATIONAXIS,
};

#endif
